from typing import Dict, List, Optional, Union, TYPE_CHECKING
from urllib.parse import urljoin, urlencode

from .base import Base
from .utils import clean_data, join_url_params
from .model3d import Model
from .file import File
from .feature import Feature

if TYPE_CHECKING:
    from . import GeoboxClient
    from .user import User
    from .task import Task
    from .attachment import Attachment
    from .aio import AsyncGeoboxClient
    from .aio.map import Map as AsyncMap


class Map(Base):
    
    BASE_ENDPOINT = 'maps/'
    
    def __init__(self, 
                 api: 'GeoboxClient',
                 uuid: str,
                 data: Optional[Dict] = {}):
        """
        Initialize a Map instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the map.
            uuid (str): The unique identifier for the map.
            data (Dict, optional): The data of the map.
        """
        self.map_layers = {
            'layers': []
        }
        super().__init__(api, uuid=uuid, data=data)


    @classmethod
    def get_maps(cls, api: 'GeoboxClient', **kwargs) -> Union[List['Map'], int]:
        """
        Get list of maps with optional filtering and pagination.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared maps. default is False.

        Returns:
            List[Map] | int: A list of Map instances or the total number of maps.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> maps = Map.get_maps(client, q="name LIKE '%My Map%'")
            or
            >>> maps = client.get_maps(q="name LIKE '%My Map%'")
        """
        params = {
           'f': 'json',
           'q': kwargs.get('q'),
           'search': kwargs.get('search'),
           'search_fields': kwargs.get('search_fields'),
           'order_by': kwargs.get('order_by'),
           'return_count': kwargs.get('return_count', False),
           'skip': kwargs.get('skip', 0),
           'limit': kwargs.get('limit', 10),
           'user_id': kwargs.get('user_id'),
           'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: Map(api, item['uuid'], item))
    

    @classmethod
    def create_map(cls, 
                   api: 'GeoboxClient', 
                   name: str, 
                   display_name: str = None, 
                   description: str = None,
                   extent: List[float] = None,
                   thumbnail: str = None,
                   style: Dict = None,
                   user_id: int = None) -> 'Map':
        """
        Create a new map.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the map.
            display_name (str, optional): The display name of the map.
            description (str, optional): The description of the map.
            extent (List[float], optional): The extent of the map.
            thumbnail (str, optional): The thumbnail of the map.
            style (Dict, optional): The style of the map.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            Map: The newly created map instance.

        Raises:
            ValidationError: If the map data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.create_map(client, name="my_map", display_name="My Map", description="This is a description of my map", extent=[10, 20, 30, 40], thumbnail="https://example.com/thumbnail.png", style={"type": "style"})
            or
            >>> map = client.create_map(name="my_map", display_name="My Map", description="This is a description of my map", extent=[10, 20, 30, 40], thumbnail="https://example.com/thumbnail.png", style={"type": "style"})
        """
        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "extent": extent,
            "thumbnail": thumbnail,
            "style": style,
            "user_id": user_id,
        }
        return super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: Map(api, item['uuid'], item))


    @classmethod
    def get_map(cls, api: 'GeoboxClient', uuid: str, user_id: int = None) -> 'Map':
        """
        Get a map by its UUID.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The UUID of the map to get.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            Map: The map object.

        Raises:
            NotFoundError: If the map with the specified UUID is not found.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> map = client.get_map(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id,
        }
        return super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: Map(api, item['uuid'], item))


    @classmethod
    def get_map_by_name(cls, api: 'GeoboxClient', name: str, user_id: int = None) -> Union['Map', None]:
        """
        Get a map by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the map to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            Map | None: returns the map if a map matches the given name, else None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map_by_name(client, name='test')
            or
            >>> map = client.get_map_by_name(name='test')
        """
        maps = cls.get_maps(api, q=f"name = '{name}'", user_id=user_id)
        if maps and maps[0].name == name:
            return maps[0]
        else:
            return None
    

    def update(self, **kwargs) -> Dict:
        """
        Update the map.

        Keyword Args:
            name (str): The name of the map.
            display_name (str): The display name of the map.
            description (str): The description of the map.
            extent (List[float]): The extent of the map.
            thumbnail (str): The thumbnail of the map.
            style (Dict): The style of the map.

        Returns:
            Dict: The updated map data.

        Raises:
            ValidationError: If the map data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.update(display_name="New Display Name")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
            "extent": kwargs.get('extent'),
            "thumbnail": kwargs.get('thumbnail'),
            "style": kwargs.get('style'),
        }
        return super()._update(self.endpoint, data)
    

    def delete(self) -> None:
        """
        Delete the map.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.delete()
        """
        super()._delete(self.endpoint)


    @property
    def style(self) -> Dict:
        """
        Get the style of the map.

        Returns:
            Dict: The style of the map.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.style
        """
        endpoint = urljoin(self.endpoint, 'style/')
        response = self.api.get(endpoint)
        return response
    
    
    @property
    def thumbnail(self) -> str:
        """
        Get the thumbnail URL of the map.

        Returns:
            str: The thumbnail of the map.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.thumbnail
        """
        return super()._thumbnail()
    

    def set_readonly(self, readonly: bool) -> None:
        """
        Set the readonly status of the map.

        Args:
            readonly (bool): The readonly status of the map.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.set_readonly(True)
        """
        data = clean_data({
            'readonly': readonly
        })
        endpoint = urljoin(self.endpoint, 'setReadonly/')
        response = self.api.post(endpoint, data, is_json=False)
        self._update_properties(response)


    def set_multiuser(self, multiuser: bool) -> None:
        """
        Set the multiuser status of the map.

        Args:
            multiuser (bool): The multiuser status of the map.

        Returns:
            None
            
        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.set_multiuser(True)
        """
        data = clean_data({
            'multiuser': multiuser
        })
        endpoint = urljoin(self.endpoint, 'setMultiuser/')
        response = self.api.post(endpoint, data, is_json=False)
        self._update_properties(response)



    def wmts(self, scale: int = None) -> str:
        """
        Get the WMTS URL of the map.

        Args:
            scale (int): The scale of the map. value are: 1, 2

        Returns:
            str: The WMTS URL of the map.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.wmts(scale=1)
        """
        endpoint = urljoin(self.api.base_url, f'{self.endpoint}wmts/')
        if scale:
            endpoint = f"{endpoint}?scale={scale}"

        if not self.api.access_token and self.api.apikey:
            endpoint = join_url_params(endpoint, {"apikey": self.api.apikey})

        return endpoint

    
    def share(self, users: List['User']) -> None:
        """
        Shares the map with specified users.

        Args:
            users (List[User]): The list of user objects to share the map with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient             
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> map.share(users=users)
        """
        super()._share(self.endpoint, users)
    

    def unshare(self, users: List['User']) -> None:
        """
        Unshares the map with specified users.

        Args:
            users (List[User]): The list of user objects to unshare the map with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient             
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> map.unshare(users=users)
        """
        super()._unshare(self.endpoint, users)


    def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['User']:
        """
        Retrieves the list of users the map is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return super()._get_shared_users(self.endpoint, params)


    def seed_cache(self, 
                   from_zoom: int = None, 
                   to_zoom: int = None, 
                   extent: List[float] = None, 
                   workers: int = 1, 
                   user_id: int = None, 
                   scale: int = None) -> List['Task']:
        """
        Seed the cache of the map.

        Args:
            from_zoom (int, optional): The zoom level to start caching from.
            to_zoom (int, optional): The zoom level to stop caching at.
            extent (List[float], optional): The extent of the map.
            workers (int, optional): The number of workers to use. default is 1.
            user_id (int, optional): Specific user. privileges required.
            scale (int, optional): The scale of the map.

        Returns:
            List[Task]: The task instance of the cache seeding operation.

        Raises:
            ValueError: If the workers is not in [1, 2, 4, 8, 12, 16, 20, 24].
            ValueError: If the cache seeding fails.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = map.seed_cache(from_zoom=0, to_zoom=10, extent=[10, 20, 30, 40], workers=1, scale=1)
        """
        data = {
            'from_zoom': from_zoom,
            'to_zoom': to_zoom,
            'extent': extent,
            'workers': workers,
            'user_id': user_id,
            'scale': scale
        }
        return super()._seed_cache(self.endpoint, data)


    def update_cache(self, 
                     from_zoom: int = None, 
                     to_zoom: int = None, 
                     extent: List[float] = None, 
                     user_id: int = None, 
                     scale: int = None) -> List['Task']:
        """
        Update the cache of the map.

        Args:
            from_zoom (int, optional): The zoom level to start caching from.
            to_zoom (int, optional): The zoom level to stop caching at.
            extent (List[float], optional): The extent of the map.
            user_id (int, optional): Specific user. privileges required.
            scale (int, optional): The scale of the map.

        Returns:
            List[Task]: The task instance of the cache updating operation.

        Raises:
            ValueError: If the cache updating fails.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.update_cache(from_zoom=0, to_zoom=10, extent=[10, 20, 30, 40], scale=1)
        """
        data = {
            'from_zoom': from_zoom,
            'to_zoom': to_zoom,
            'extent': extent,
            'user_id': user_id,
            'scale': scale
        }
        return super()._update_cache(self.endpoint, data)


    def clear_cache(self) -> None:
        """
        Clear the cache of the map.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.clear_cache()
        """
        return super()._clear_cache(self.endpoint)


    @property
    def cache_size(self) -> int:
        """
        Get the size of the cache of the map.

        Returns:
            int: The size of the cache of the map.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.cache_size
        """
        return super()._cache_size(self.endpoint)


    @property
    def settings(self) -> Dict:
        """
        Get the settings of the map

        Returns:
            Dict: the settings of the map.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(uuid="12345678-1234-5678-1234-567812345678")
            >>> map.settings
        """
        return self.json.get('settings', {
            'general_settings': {},
            'edit_settings': {},
            'snap_settings': {},
            'controls': [],
            'search_settings': {},
            'marker_settings': {},
            'terrain_settings': {},
            'grid_settings': {},
            'view_settings': {},
            'toc_settings': []
        })
    

    def update_settings(self, settings: Dict) -> Dict:
        """
        Update the settings

        settings (Dict): settings dictionary

        Returns:
            Dict: updated settings

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> map1 = client.get_map(uuid="12345678-1234-5678-1234-567812345678")
            >>> map2 = client.get_map(uuid="12345678-1234-5678-1234-567812345678")
            >>> map1.update_settings(map2.settings)
        """
        return super()._set_settings(self.endpoint, settings)    


    def set_settings(self, **kwargs) -> Dict:
        """
        Set the settings of the map using keywords

        Keyword Args:
            map_unit (str): 'latlng' | 'utm'.
            base_map (str): 'OSM' | 'google' | 'blank'.
            flash_color (str): 'rgb(255,0,0)' (rgb color or rgba color or hex color ).
            highlight_color (str): 'rgb(255,0,0)' (rgb color or rgba color or hex color ).
            selection_color (str): 'rgb(255,0,0)' (rgb color or rgba color or hex color ).
            selectable_layers (str): 'ALL' | null | Comma separated list of layers.
            calendar_type (str): The type of the calendar.
            edit_settings (dict): The settings of the edit.
            snap_tolerance (int): number of pixels for snap tolerance.
            snap_unit (str): pixels.
            snap_mode (str): 'both' | 'edge' | 'vertex'.
            snap_cache (int): number of total features for snap cache.
            controls (List[str]): The controls of the map.
            search_mode (str): 'both' | 'markers' | 'layers'.
            search_layers (str): 'ALL' | null | Comma separated list of layers.
            geosearch (bool): The geosearch of the map.
            remove_unused_tags (bool): The remove unused tags of the map.
            terrain_layer (str): The terrain layer of the map.
            exaggeration (int): The exaggeration of the terrain.
            enable_grid (bool): The enable grid of the map.
            grid_unit (str): The unit of the grid.
            grid_width (int): The width of the grid.
            grid_height (int): The height of the grid.
            grid_minzoom (int): The minzoom of the grid.
            grid_maxzoom (int): The maxzoom of the grid.
            bearing (int): The bearing of the map.
            pitch (int): The pitch of the map.
            center (List[float]): The center of the map.
            zoom (int): The zoom of the map.
            toc_settings (List): The settings of the toc.
            custom_basemaps (List[str]): The custom basemaps of the map.
            show_maptip_on (str): 'ALL' | null | Comma separated list of layers.
            snappable_layers (str): 'ALL' | null | Comma separated list of layers.

        Returns:
            Dict: The response of the API.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(uuid="12345678-1234-5678-1234-567812345678")
            >>> map.set_settings(zoom=10)
        """
        general_settings = {'map_unit', 'base_map', 'flash_color', 'highlight_color', 
                            'selection_color', 'selectable_layers', 'calendar_type', 'custom_basemaps', 'show_maptip_on'}
        edit_settings = {'editable_layers', 'target_layer'}
        snap_settings = {'snap_tolerance', 'snap_unit', 'snap_mode', 'snap_cache', 'snappable_layers'}
        search_settings = {'search_mode', 'search_layers', 'geosearch'}
        marker_settings = {'remove_unused_tags'}
        terrain_settings = {'terrain_layer', 'exaggeration'}
        grid_settings = {'enable_grid', 'grid_unit', 'grid_width', 'grid_height', 'grid_minzoom', 'grid_maxzoom'}
        view_settings = {'bearing', 'pitch', 'center', 'zoom'}

        settings = self.settings

        for key, value in kwargs.items():
            if key in general_settings:
                settings['general_settings'][key] = value
            elif key in edit_settings:
                settings['edit_settings'][key] = value
            elif key in snap_settings:
                settings['snap_settings'][key] = value
            elif key == 'controls':
                settings['controls'] = value
            elif key in search_settings:
                settings['search_settings'][key] = value
            elif key in marker_settings:
                settings['marker_settings'][key] = value
            elif key in terrain_settings:
                settings['terrain_settings'][key] = value
            elif key in grid_settings:
                settings['grid_settings'][key] = value
            elif key in view_settings:
                settings['view_settings'][key] = value
            elif key == 'toc_settings':
                settings['toc_settings'] = value

        return super()._set_settings(self.endpoint, settings)    


    def get_markers(self) -> Dict:
        """
        Get the markers of the map.

        Returns:
            Dict: The markers of the map.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.get_markers()
        """
        endpoint = urljoin(self.endpoint, 'markers/')
        return self.api.get(endpoint)
    

    def set_markers(self, data: Dict) -> Dict:
        """
        Set the markers of the map.

        Args:
            data (dict): The data of the markers.

        Returns:
            Dict: The response of the API.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> data = {
            ...     'tags': {
            ...         '#general': {
            ...             'color': '#ff0000',
            ...         }
            ...     },
            ...     'locations': [
            ...         {
            ...             'id': 1,
            ...             'tag': '#general',
            ...             'name': 'test',
            ...             'geometry': [
            ...                 51.13162784422988,
            ...                 35.766603814763045
            ...             ],
            ...             'description': 'string'
            ...         }
            ...     ]
            ... }
            >>> map.set_markers(data)
        """
        endpoint = urljoin(self.endpoint, 'markers/')
        response = self.api.put(endpoint, data)
        return response
    

    def get_models(self, json=False) -> Union[List['Model'], Dict]:
        """
        Get the map models.

        Args:
            json (bool, optional): If True, return the response as a dictionary.

        Returns:
            List[Model] | Dict: map models objects or the response.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.get_models(json=True)
        """
        endpoint = urljoin(self.endpoint, 'models/')
        response = self.api.get(endpoint)
        if not response or json:
            return response
        else:
            return [Model(self.api, model['obj'], model) for model in response['objects'] if response.get('objects')]
    

    def set_models(self, data: Dict) -> List['Model']:
        """
        Set multiple models on the map.

        Args:
            data (Dict): the data of the models and their location on the map. check the example for the data structure.

        Returns:
            List[Model]: the map models objects

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map inport Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(uuid="12345678-1234-5678-1234-567812345678")
            >>> data = {'objects': [
            ...     {
            ...         "name": "transmission_tower",
            ...         "alias": None,
            ...         "desc": None,
            ...         "obj": "12345678-1234-5678-1234-567812345678",
            ...         "loc": [53.1859045261684, 33.37762747390032, 0.0],
            ...         "rotation": [0.0, 0.0, 0.0],
            ...         "scale": 1.0,
            ...         "min_zoom": 0,
            ...         "max_zoom": 22
            ...     }
            ... ]}
            >>> map.set_models(data)
        """
        endpoint = urljoin(self.endpoint, 'models/')
        response = self.api.put(endpoint, data)
        return [Model(self.api, model['obj'], model) for model in response['objects'] if response.get('objects')]


    def add_model(self, 
                   model: 'Model', 
                   location: List[float], 
                   rotation: List[float] = [0.0, 0.0, 0.0],
                   scale: float = 1.0,
                   min_zoom: int = 0,
                   max_zoom: int = 22,
                   alias: str = None,
                   description: str = None) -> List['Model']:
        """
        Add a model the map.

        Args:
            model (Model): The model object.
            location (List[float]): location of the model on the map. a list with three float values.
            rotation (List[float], optional): rotation of the model on the map. a list with three float vlaues. default is [0.0, 0.0, 0.0].
            scale (float, optional): the scale of the model on the map.
            min_zoom (int, optional): minimum zoom level.
            max_zoom (int, optional): maximum zoom level.
            alias (str, optional): alias of the model on the map.
            description (str, optional): the description of the model on the map.

        Returns:
            List['Model']: The map model objects

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> model = client.get_model(uuid="12345678-1234-5678-1234-567812345678")
            >>> map.add_model(model=model,
            ...                 location=[53.53, 33.33, 0.0],
            ...                 rotation=[0.0, 0.0, 0.0],
            ...                 scale=1.0,
            ...                 min_zoom=0,
            ...                 max_zoom=22,
            ...                 alias=None,
            ...                 description=None)
        """
        data = self.get_models(json=True)
        if data and data.get('objects') and isinstance(data['objects'], list):
            data.get('objects').append({
                        'name': model.name,
                        'alias': alias,
                        'desc': description,
                        'obj': model.uuid,
                        'loc': location,
                        'rotation': rotation,
                        'scale': scale,
                        'min_zoom': min_zoom,
                        'max_zoom': max_zoom
                    })
        else:
            data = {'objects':[
                    {
                        'name': model.name,
                        'alias': alias,
                        'desc': description,
                        'obj': model.uuid,
                        'loc': location,
                        'rotation': rotation,
                        'scale': scale,
                        'min_zoom': min_zoom,
                        'max_zoom': max_zoom
                    }
                ]
            }

        endpoint = urljoin(self.endpoint, 'models/')
        response = self.api.put(endpoint, data)
        return [Model(self.api, model['obj'], model) for model in response['objects']]


    def image_tile_url(self, x: str = '{x}', y: str = '{y}', z: str = '{z}', format='.png') -> str:
        """
        Get map image tile url

        Returns:
            str: the image tile url

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.image_tile_url()
            >>> map.image_tile_url(x=1, y=2, z=3, format='.pbf')
        """
        endpoint = f"{self.api.base_url}{self.endpoint}tiles/{z}/{x}/{y}{format}"

        if not self.api.access_token and self.api.apikey:
            endpoint = f'{endpoint}?apikey={self.api.apikey}'

        return endpoint
    

    def export_map_to_image(self, bbox: List, width: int, height: int) -> 'Task':
        """
        Export the map to image

        Args:
            bbox (List): e.g. [50.275, 35.1195, 51.4459, 36.0416]
            width (int): minimum: 10, maximum: 10000
            height (int): minimum: 10, maximum: 10000

        Returns:
            Task: the task object

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = map.export_map_to_image(bbox=[50.275, 35.1195, 51.4459, 36.0416],
            ...                                 width=1024,
            ...                                 height=1024)
        """
        data = clean_data({
            'uuid': self.uuid,
            'bbox': bbox,
            'width': width,
            'height': height
        })
        query_string = urlencode(data)
        endpoint = urljoin(self.endpoint, 'export/')
        endpoint = f"{endpoint}?{query_string}"
        response = self.api.post(endpoint)
        return self.api.get_task(response['task_id'])


    def get_attachments(self, **kwargs) -> List['Attachment']:
        """
        Get the resouces attachments

        Keyword Args:
            element_id (str): the id of the element with attachment.
            search (str): search term for keyword-based searching among all textual fields.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            return_count (bool): Whether to return total count. default is False.

        Returns:
            List[Attachment] | int: A list of attachments instances or the total number of attachments.

        Raises:
            TypeError: if the resource type is not supported

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> map.get_attachments()
        """
        from .attachment import Attachment

        return Attachment.get_attachments(self.api, resource=self, **kwargs)
    
    
    def create_attachment(self, 
                            name: str, 
                            loc_x: int,
                            loc_y: int,
                            file: 'File',
                            feature: 'Feature' = None,
                            display_name: str = None, 
                            description: str = None) -> 'Attachment':
        """
        Create a new Attachment.

        Args:
            name (str): The name of the scene.
            loc_x (int): x parameter of the attachment location.
            loc_y (int): y parameter of the attachment location.
            file (File): the file object.
            feature (Feature, optional): the feature object.
            display_name (str, optional): The display name of the scene.
            description (str, optional): The description of the scene.

        Returns:
            Attachment: The newly created Attachment instance.

        Raises:
            ValidationError: If the Attachment data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file = client.get_files()[0]
            >>> map.create_attachment(name='test', loc_x=10, loc_y=10, file=file)
        """
        from .attachment import Attachment

        return Attachment.create_attachment(self.api,
                                            name=name,
                                            loc_x=loc_x,
                                            loc_y=loc_y,
                                            resource=self,
                                            file=file,
                                            feature=feature,
                                            display_name=display_name,
                                            description=description)
    

    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncMap':
        """
        Switch to async version of the map instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncMap: the async instance of the map.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.map import Map
            >>> client = GeoboxClient()
            >>> map = Map.get_map(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_map = map.to_async(async_client)
        """
        from .aio.map import AsyncMap

        return AsyncMap(api=async_client, uuid=self.uuid, data=self.data)


from urllib.parse import urljoin
from typing import Optional, Dict, List, Union, TYPE_CHECKING
import os
import mimetypes
from geobox.exception import ValidationError
import requests
import sys

from .base import Base
from .enums import FileFormat, PublishFileType, InputGeomType, FileType
from .utils import clean_data, get_unique_filename, get_save_path
from .task import Task
from .feature import Feature

if TYPE_CHECKING:
    from . import GeoboxClient 
    from .user import User
    from .aio import AsyncGeoboxClient
    from .aio.file import AsyncFile


class File(Base):

    BASE_ENDPOINT: str = 'files/'

    def __init__(self, 
                api: 'GeoboxClient', 
                uuid: str, 
                data: Optional[Dict] = {}):
        """
        Constructs all the necessary attributes for the File object.

        Args:
            api (GeoboxClient): The GeoboxClient instance.
            uuid (str): The UUID of the file.
            data (Dict, optional): The data of the file.
        """
        super().__init__(api, uuid=uuid, data=data)


    def __repr__(self) -> str:
        """
        Return a string representation of the File object.

        Returns:
            str: A string representation of the File object.
        """
        return f"File(uuid={self.uuid}, file_name={self.name}, file_type={self.file_type.value})"


    @property
    def layers(self) -> List[Dict]:
        """
        Get the layers of the file.

        Returns:
            List[Dict]: The layers of the file.
        
        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.layers
        """
        return self.data.get('layers', {}).get('layers', [])


    @property
    def file_type(self) -> 'FileType':
        """
        Get the file type

        Returns:
            FileType: the file type enumeration

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.file_type
        """
        return FileType(self.data.get('file_type'))


    @classmethod
    def upload_file(cls, api: 'GeoboxClient', path: str, user_id: int = None, scan_archive: bool = True) -> 'File':
        """
        Upload a file to the GeoBox API.

        Args:
            api (GeoboxClient): The GeoboxClient instance.
            path (str): The path to the file to upload.
            user_id (int, optional): specific user. privileges required.
            scan_archive (bool, optional): Whether to scan the archive for layers. default: True

        Returns:
            File: The uploaded file instance.

        Raises:
            ValueError: If the file type is invalid.
            FileNotFoundError: If the file does not exist.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.upload_file(client, path='path/to/file.shp')
            or
            >>> file = client.upload_file(path='path/to/file.shp')
        """
        # Check if the file exists
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {path}")

        # Check if the file type is valid
        FileFormat(os.path.splitext(path)[1])

        data = clean_data({
            "user_id": user_id,
            "scan_archive": scan_archive
        })
        
        endpoint = cls.BASE_ENDPOINT
        with open(path, 'rb') as f:
            files = {'file': f}
            file_data = api.post(endpoint, data, is_json=False, files=files)
        return cls(api, file_data['uuid'], file_data)


    @classmethod
    def get_files(cls, api:'GeoboxClient', **kwargs) -> Union[List['File'], int]:
        """
        Retrieves a list of files.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): if true, the total number of results will be returned. default is False.
            skip (int): number of results to skip. default is 0.
            limit (int): number of results to return. default is 10.
            user_id (int): filter by user id.
            shared (bool): Whether to return shared files. default is False.
            
        Returns:
            List[File] | int: A list of File objects or the total number of results.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File 
            >>> client = GeoboxClient()
            >>> files = File.get_files(client, search_fields='name', search='GIS', order_by='name', skip=10, limit=10)
            or
            >>> files = client.get_files(search_fields='name', search='GIS', order_by='name', skip=10, limit=10)
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: cls(api, item['uuid'], item))


    @classmethod
    def get_file(cls, api: 'GeoboxClient', uuid: str, user_id: int = None) -> 'File':
        """
        Retrieves a file by its UUID.

        Args:
            api (Api): The GeoboxClient instance.
            uuid (str): The UUID of the file.
            user_id (int, optional): specific user. privileges required.

        Returns:
            File: The retrieved file instance.

        Raises:
            NotFoundError: If the file with the specified UUID is not found.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> file = client.get_file(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return super()._get_detail(api, cls.BASE_ENDPOINT, f'{uuid}/info', params, factory_func=lambda api, item: File(api, item['uuid'], item))


    @classmethod
    def get_files_by_name(cls, api: 'GeoboxClient', name: str, user_id: int = None) -> List['File']:
        """
        Get files by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the file to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            List[File]: returns files that matches the given name

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> files = File.get_files_by_name(client, name='test')
            or
            >>> files = client.get_files_by_name(name='test')
        """
        return cls.get_files(api, q=f"name = '{name}'", user_id=user_id)


    def _get_file_name(self, response: requests.Response) -> str:
        """
        Get the file name from the response.

        Args:
            response (requests.Response): The response of the request.

        Returns:
            str: The file name 
        """
        if 'Content-Disposition' in response.headers and 'filename=' in response.headers['Content-Disposition']:
            file_name = response.headers['Content-Disposition'].split('filename=')[-1].strip().strip('"')

        else:
            content_type = response.headers.get("Content-Type", "")
            file_name = f'{self.name}.{mimetypes.guess_extension(content_type.split(";")[0])}'

        return file_name


    def _create_progress_bar(self) -> 'tqdm':
        """Creates a progress bar for the task."""
        try:
            from tqdm.auto import tqdm
        except ImportError:
            from .api import logger
            logger.warning("[tqdm] extra is required to show the progress bar. install with: pip insatll geobox[tqdm]")
            return None

        return tqdm(unit="B", 
                        total=int(self.size), 
                        file=sys.stdout,
                        dynamic_ncols=True,
                        desc="Downloading",
                        unit_scale=True,
                        unit_divisor=1024, 
                        ascii=True
                )


    def download(self, save_path: str = None, progress_bar: bool = True, file_name: str = None, overwrite: bool = False) -> str:
        """
        Download a file and save it to the specified path.
        
        Args:
            save_path (str, optional): Path where the file should be saved. 
                                    If not provided, it saves to the current working directory
                                    using the original filename and appropriate extension.
            progress_bar (bool, optional): Whether to show a progress bar. default: True
            file_name (str, optional): the downloaded file name.
            overwrite (bool, optional): whether to overwrite the downloaded file if it exists on the save path. default is False.
        
        Returns:
            str: Path where the file was saved
            
        Raises:
            ValueError: If uuid is not set
            OSError: If there are issues with file operations

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.download(save_path='path/to/save/')
        """
        if not self.uuid:
            raise ValueError("File UUID is required to download the file")
        
        save_path = get_save_path(save_path)
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        
        with self.api.get(f"{self.endpoint}download/", stream=True) as response:
            file_name = self._get_file_name(response) if not file_name else file_name
            full_path = f"{save_path}/{file_name}"
            if os.path.exists(full_path) and not overwrite:
                full_path = get_unique_filename(save_path, file_name)
            with open(full_path, 'wb') as f:
                pbar = self._create_progress_bar() if progress_bar else None
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if pbar:
                        pbar.update(len(chunk))
                        pbar.refresh()
                if pbar:
                    pbar.close()

        return os.path.abspath(full_path)


    def delete(self) -> None:
        """
        Deletes the file.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.delete()
        """
        return super()._delete(self.endpoint)


    def publish(self, 
                    name: str, 
                    publish_as: 'PublishFileType' = None, 
                    input_geom_type: 'InputGeomType' = None, 
                    input_layer: str = None, 
                    input_dataset: str = None, 
                    user_id: int = None, 
                    input_srid: int = Feature.BASE_SRID, 
                    file_encoding: str = "UTF-8", 
                    replace_domain_codes_by_values: bool = False, 
                    report_errors: bool = True, 
                    as_terrain: bool = False) -> 'Task':
        """
        Publishes a file as a layer.

        Args:
            name (str): The name of the layer.
            publish_as (PublishFileType, optional): The type of layer to publish as.
            input_geom_type (InputGeomType, optional): The geometry type of the layer.
            input_layer (str, optional): The name of the input layer.
            input_dataset (str, optional): The name of the input dataset.
            user_id (int, optional): Specific user. privileges required.
            input_srid (int, optional): The SRID of the layer. default is: 3857
            file_encoding (str, optional): The encoding of the file. default is "utf-8".
            replace_domain_codes_by_values (bool, optional): Whether to replace domain codes by values. default is False.
            report_errors (bool, optional): Whether to report errors. default is True.
            as_terrain (bool, optional): Whether to publish as terrain. default is False.

        Returns:
            Task: The task object.

        Raises:
            ValueError: If the publish_as is not a valid PublishFileType.
            ValidationError: if the zipped file doesn't have any layers to publish.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.publish(publish_as=PublishFileType.VECTOR, 
            ...                     layer_name='my_layer', 
            ...                     input_geom_type=InputGeomType.POINT, 
            ...                     input_layer='layer1', 
            ...                     input_dataset='dataset1', 
            ...                     input_srid=4326, 
            ...                     file_encoding='UTF-8')
        """
        if not publish_as:
            # checks the file format or file first layer format to dynamically set the publish_as
            if self.file_type.value in ['GeoJSON', 'GPKG', 'DXF', 'GPX', 'Shapefile', 'KML', 'CSV', 'FileGDB'] or \
                (self.file_type.value in ['Complex'] and self.layers and \
                    FileType(self.layers[0]['format']).value in ['GeoJSON', 'GPKG', 'DXF', 'GPX', 'Shapefile', 'KML', 'CSV', 'FileGDB']):
                publish_as = PublishFileType.VECTOR

            elif self.file_type.value in ['GeoTIFF'] or \
                    (self.file_type.value in ['Complex'] and self.layers and \
                        FileType(self.layers[0]['format']).value in ['GeoTIFF']):
                publish_as = PublishFileType.RASTER

            elif self.file_type.value in ['GLB'] or \
                    (self.file_type.value in ['Complex'] and self.layers and \
                        FileType(self.layers[0]['format']).value in ['GLB']):
                publish_as = PublishFileType.MODEL3D

            elif self.file_type.value in ['ThreedTiles']:
                publish_as = PublishFileType.Tiles3D

            else:
                raise ValidationError('Unknown format')

        data = clean_data({
            "publish_as": publish_as.value if isinstance(publish_as, PublishFileType) else publish_as,
            "layer_name": name,
            "input_layer": self.layers[0]['layer'] if not input_layer and self.layers else input_layer,
            "input_geom_type": input_geom_type.value if isinstance(input_geom_type, InputGeomType) else input_geom_type,
            "replace_domain_codes_by_values": replace_domain_codes_by_values,
            "input_dataset": self.layers[0]['dataset'] if not input_layer and self.layers else input_dataset,
            "user_id": user_id,
            "input_srid": input_srid,
            "file_encoding": file_encoding,
            "report_errors": report_errors,
            "as_terrain": as_terrain
        })
        endpoint = urljoin(self.endpoint, 'publish/')
        response = self.api.post(endpoint, data, is_json=False)
        task = Task.get_task(self.api, response.get('task_id'))
        return task


    def share(self, users: List['User']) -> None:
        """
        Shares the file with specified users.

        Args:
            users (List[User]): The list of users objects to share the file with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> file.share(users=users)
        """
        super()._share(self.endpoint, users)
    

    def unshare(self, users: List['User']) -> None:
        """
        Unshares the file with specified users.

        Args:
            users (List[User]): The list of users objects to unshare the file with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> file.unshare(users=users)
        """
        super()._unshare(self.endpoint, users)


    def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['User']:
        """
        Retrieves the list of users the file is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return super()._get_shared_users(self.endpoint, params)


    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncFile':
        """
        Switch to async version of the file instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncFile: the async instance of the file.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.file import File
            >>> client = GeoboxClient()
            >>> file = File.get_file(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_file = file.to_async(async_client)
        """
        from .aio.file import AsyncFile

        return AsyncFile(api=async_client, uuid=self.uuid, data=self.data)


from urllib.parse import urlencode
from typing import Dict, List, Union, TYPE_CHECKING
from datetime import datetime

from .base import AsyncBase
from .user import AsyncUser
from .apikey import AsyncApiKey
from ..utils import clean_data
from ..enums import UsageScale, UsageParam

if TYPE_CHECKING:
    from . import AsyncGeoboxClient
    from ..api import GeoboxClient
    from ..usage import Usage


class AsyncUsage(AsyncBase):

    BASE_ENDPOINT = 'usage/'

    def __init__(self,
        api: 'AsyncGeoboxClient',
        user: 'AsyncUser'):
        """
        Constructs the necessary attributes for the Usage object.

        Args:
            api (AsyncGeoboxClient): The API instance.
            user (User): the user usage object.
        """
        self.api = api
        self.user = user


    def __repr__(self) -> str:
        """
        Return a string representation of the Usage object.

        Returns:
            str: A string representation of the Usage object.
        """
        return f"AsyncUsage(user={self.user})"


    @classmethod
    async def get_api_usage(cls, 
        api: 'AsyncGeoboxClient', 
        resource: Union['AsyncUser', 'AsyncApiKey'], 
        scale: 'UsageScale',
        param: 'UsageParam',
        from_date: 'datetime' = None,
        to_date: 'datetime' = None,
        days_before_now: int = None,
        limit: int = None) -> List:
        """
        [async] Get the api usage of a user

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            resource (AsyncUser | AsyncApiKey): User or ApiKey object.
            scale (UsageScale): the scale of the report.
            param (UsageParam): traffic or calls.
            from_date (datetime, optional): datetime object in this format: "%Y-%m-%dT%H:%M:%S". 
            to_date (datetime, optional): datetime object in this format: "%Y-%m-%dT%H:%M:%S". 
            days_before_now (int, optional): number of days befor now.
            limit (int, optional): Number of items to return. default is 10.

        Raises:
            ValueError: one of days_before_now or from_date/to_date parameters must have value
            ValueError: resource must be a 'user' or 'apikey' object

        Returns:
            List: usage report

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.usage import AsyncUsage
            >>> async with AsyncGeoboxClient() as client:
            >>>     user = await client.get_user() # gets current user
            >>>     usage = await AsyncUsage.get_api_usage(client, 
            ...                                         resource=user, 
            ...                                         scale=UsageScale.Day, 
            ...                                         param=UsageParam.Calls, 
            ...                                         days_before_now=5)
            or  
            >>>     usage = await client.get_api_usage(resource=user, 
            ...                                         scale=UsageScale.Day, 
            ...                                         param=UsageParam.Calls, 
            ...                                         days_before_now=5)
        """
        if not(from_date and to_date) and not days_before_now:
            raise ValueError("one of days_before_now or from_date/to_date parameters must have value")

        params = {}
        if isinstance(resource, AsyncUser):
            params['eid'] = resource.user_id
        elif isinstance(resource, AsyncApiKey):
            params['eid'] = resource.key
        else:
            raise ValueError("resource must be a 'AsyncUser' or 'AsyncApikey' object")

        params = clean_data({
            **params,
            'scale': scale.value if scale else None,
            'param': param.value if param else None,
            'from_date': from_date.strftime("%Y-%m-%dT%H:%M:%S.%f") if from_date else None,
            'to_date': to_date.strftime("%Y-%m-%dT%H:%M:%S.%f") if to_date else None,
            'days_before_now': days_before_now,
            'limit': limit
        })
        query_strings = urlencode(params)
        endpoint = f"{cls.BASE_ENDPOINT}api?{query_strings}"

        return await api.get(endpoint)


    @classmethod
    async def get_process_usage(cls, 
        api: 'AsyncGeoboxClient',
        user_id: int = None, 
        from_date: datetime = None, 
        to_date: datetime = None, 
        days_before_now: int = None) -> float:
        """
        [async] Get process usage of a user in seconds

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            user_id (int, optional): the id of the user. leave blank to get the current user report.
            from_date (datetime, optional): datetime object in this format: "%Y-%m-%dT%H:%M:%S". 
            to_date (datetime, optional): datetime object in this format: "%Y-%m-%dT%H:%M:%S". 
            days_before_now (int, optional): number of days befor now.

        Raises:
            ValueError: one of days_before_now or from_date/to_date parameters must have value

        Returns:
            float: process usage of a user in seconds

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.usage import AsyncUsage
            >>> async with AsyncGeoboxClient() as client:
            >>>     process_usage = await AsyncUsage.get_process_usage(client, days_before_now=5)
            or  
            >>>     process_usage = await client.get_process_usage(days_before_now=5)
        """
        if not(from_date and to_date) and not days_before_now:
            raise ValueError("one of days_before_now or from_date/to_date parameters must have value")

        params = clean_data({
            'user_id': user_id if user_id else None,
            'from_date': from_date.strftime("%Y-%m-%dT%H:%M:%S.%f") if from_date else None,
            'to_date': to_date.strftime("%Y-%m-%dT%H:%M:%S.%f") if to_date else None,
            'days_before_now': days_before_now
        })
        query_strings = urlencode(params)
        endpoint = f"{cls.BASE_ENDPOINT}process?{query_strings}"
        return await api.get(endpoint)


    @classmethod
    async def get_usage_summary(cls, api: 'AsyncGeoboxClient', user_id: int = None) -> Dict:
        """
        [async] Get the usage summary of a user

        Args:
            api (AsyncGeoboxClient): The API instance.
            user_id (int, optional): the id of the user. leave blank to get the current user report.

        Returns:
            Dict: the usage summary of the users

        Returns:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.usage import AsyncUsage
            >>> async with AsyncGeoboxClient() as client:
            >>>     usage_summary = await AsyncUsage.get_usage_summary(client)
            or  
            >>>     usage_summary = await client.get_usage_summary()
        """
        params = clean_data({
            'user_id': user_id if user_id else None
        })
        query_strings = urlencode(params)
        endpoint = f"{cls.BASE_ENDPOINT}summary?{query_strings}"
        return await api.get(endpoint)


    @classmethod
    async def update_usage(cls, api: 'AsyncGeoboxClient', user_id: int = None) -> Dict:
        """
        [async] Update usage of a user

        Args:
            api (AsyncGeoboxClient): The API instance.
            user_id (int, optional): the id of the user. leave blank to get the current user report.
            
        Returns:
            Dict: the updated data

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.usage import AsyncUsage
            >>> async with AsyncGeoboxClient() as client:
            >>>     await AsyncUsage.update_usage(client)
            or  
            >>>     await client.update_usage()
        """
        data = clean_data({
            'user_id': user_id if user_id else None
        })
        endpoint = f"{cls.BASE_ENDPOINT}update"
        return await api.post(endpoint, payload=data)


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Usage':
        """
        Switch to sync version of the usage instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Usage: the sync instance of the usage.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.usage import AsyncUsage
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     user = await async_client.get_user()
            >>>     usage = await AsyncUsage.get_api_usage(async_client, 
            ...                                         resource=user, 
            ...                                         scale=UsageScale.Day, 
            ...                                         param=UsageParam.Calls, 
            ...                                         days_before_now=5)
            >>>     sync_usage = await usage.to_sync(client)
        """
        from ..usage import Usage

        user = self.user.to_sync(sync_client)
        return Usage(api=sync_client, user=user)


from typing import List, Optional, Dict, List, Union, TYPE_CHECKING

from .vectorlayer import VectorLayer, LayerType, FileOutputFormat
from .feature import Feature

if TYPE_CHECKING:
    from . import GeoboxClient
    from .field import Field
    from .user import User
    from .enums import InputGeomType
    from .task import Task
    from .file import File
    from .attachment import Attachment
    from .aio import AsyncGeoboxClient
    from .aio.view import AsyncVectorLayerView


class VectorLayerView(VectorLayer):

    BASE_ENDPOINT = 'vectorLayerViews/'

    def __init__(self, 
                 api: 'GeoboxClient', 
                 uuid: str,
                 layer_type: 'LayerType', 
                 data: Optional[Dict] = {}) -> None:
        """
        Initialize a VectorLayerView instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The UUID of the vector layer view.
            layer_type (LayerType): The type of the vector layer view.
            data (Dict, optional): The data of the vector layer view.
        """
        super().__init__(api, uuid, layer_type, data)


    def __repr__(self) -> str:
        """
        Return a string representation of the VectorLayerView instance.

        Returns:
            str: A string representation of the VectorLayerView instance.
        """
        return f"VectorLayerView(uuid={self.uuid}, name={self.name}, layer_type={self.layer_type})"
    

    @property
    def vector_layer(self) -> 'VectorLayer':
        """
        Get the vector layer.

        Returns:
            VectorLayer: The vector layer.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid='e21e085a-8d30-407d-a740-ca9be9122c42')
            >>> view.vector_layer
        """
        return VectorLayer(self.api, self.data['vector_layer']['uuid'], LayerType(self.data['vector_layer']['layer_type']), self.data['vector_layer'])


    @classmethod
    def get_views(cls, api: 'GeoboxClient', **kwargs) -> Union[List['VectorLayerView'], int]:
        """
        Get vector layer views.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            layer_id(int): The id of the source vector layer.
            include_settings(bool): Whether to include the settings of the layer. default is False.
            q(str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search(str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields(str): Comma separated list of fields for searching.
            order_by(str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count(bool): Whether to return the count of the layer views. default is False.
            skip(int): The number of views to skip. minimum is 0.
            limit(int): The maximum number of views to return. minimum is 1. default is 10.
            user_id(int): Specific user. privileges required.
            shared(bool): Whether to return shared views. default is False.

        Returns:
            list[VectorLayerView] | int: A list of VectorLayerView instances or the layer views count if return_count is True.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> views = VectorLayerView.get_views(client,
                                                    layer_id=1,
                                                    include_settings=True,
                                                    search="test",
                                                    search_fields="name",
                                                    order_by="name A",
                                                    return_count=False,
                                                    skip=0,
                                                    limit=10,
                                                    shared=True)
            or
            >>> views = client.get_views(layer_id=1,
                                            include_settings=True,
                                            search="test",
                                            search_fields="name",
                                            order_by="name A",
                                            return_count=False,
                                            skip=0,
                                            limit=10,
                                            shared=True)
        """
        params = {
            'f': 'json',
            'include_settings': kwargs.get('include_settings', False),
            'temporary': kwargs.get('temporary', False),
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api=api, 
                                 endpoint=cls.BASE_ENDPOINT,
                                 params=params, 
                                 factory_func=lambda api, item: VectorLayerView(api, item['uuid'], LayerType(item['layer_type']), item))


    @classmethod
    def get_views_by_ids(cls, api: 'GeoboxClient', ids: List[int], user_id: int = None, include_settings: bool = False) -> List['VectorLayerView']:
        """
        Get vector layer views by their IDs.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            ids (List[int]): list of comma separated layer ids to be returned. e.g. 1, 2, 3
            user_id (int, optional): specific user. privileges required.
            include_settings (bool, optional): Whether to include the settings of the vector layer views. default is False.

        Returns:
            List[VectorLayerView]: A list of VectorLayerView instances.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> views = VectorLayerView.get_views_by_ids(client, ids=[1,2,3])
            or
            >>> views = client.get_views_by_ids(ids=[1,2,3])
        """
        params = {
            'ids': ids,
            'user_id': user_id,
            'include_settings': include_settings
        }
        return super()._get_list_by_ids(api=api, 
                                        endpoint=f'{cls.BASE_ENDPOINT}get-layers/', 
                                        params=params, 
                                        factory_func=lambda api, item: VectorLayerView(api, item['uuid'], LayerType(item['layer_type']), item))


    @classmethod
    def get_view(cls, api: 'GeoboxClient', uuid: str, user_id: int = None) -> 'VectorLayerView':
        """
        Get a specific vector layer view by its UUID.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The UUID of the vector layer view.
            user_id (int, optional): Specific user. privileges required.

        Returns:    
            VectorLayerView: A VectorLayerView instance.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> view = client.get_view(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return super()._get_detail(api=api,
                                   endpoint=cls.BASE_ENDPOINT,
                                   uuid=uuid, 
                                   params=params, 
                                   factory_func=lambda api, item: VectorLayerView(api, item['uuid'], LayerType(item['layer_type']), item))


    @classmethod
    def get_view_by_name(cls, api: 'GeoboxClient', name: str, user_id: int = None) -> Union['VectorLayerView', None]:
        """
        Get a view by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the view to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            VectorLayerView | None: returns the view if a view matches the given name, else None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view_by_name(client, name='test')
            or
            >>> view = client.get_view_by_name(name='test')
        """
        views = cls.get_views(api, q=f"name = '{name}'", user_id=user_id)
        if views and views[0].name == name:
            return views[0]
        else:
            return None


    def update(self, **kwargs) -> Dict:
        """
        Update the vector layer view.

        Keyword Args:
            name (str): The name of the vector layer view.
            display_name (str): The display name of the vector layer view.
            description (str): The description of the vector layer view.

        Returns:
            Dict: The updated vector layer view.

        Raises:
            ValidationError: If the update data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.update(name="new_name")
            >>> view.update(display_name="new_display_name")
            >>> view.update(description="new_description")
        """
        return super().update(name=kwargs.get('name'), display_name=kwargs.get('display_name'), description=kwargs.get('description'))
    

    def delete(self) -> None:
        """
        Delete the vector layer view.
        
        Returns:
            None
            
        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.delete()
        """        
        return super()._delete(self.endpoint)


    def share(self, users: List['User']) -> None:
        """
        Shares the view with specified users.

        Args:
            users (List[User]): The list of user IDs to share the view with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> view.share(users=users)
        """
        super()._share(self.endpoint, users)
    

    def unshare(self, users: List['User']) -> None:
        """
        Unshares the view with specified users.

        Args:
            users (List[User]): The list of user IDs to unshare the view with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> view.unshare(users=users)
        """
        super()._unshare(self.endpoint, users)


    def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['User']:
        """
        Retrieves the list of users the view is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return super()._get_shared_users(self.endpoint, params)


    def get_fields(self) -> List['Field']:
        """
        Get all fields in the vector layer view.
        
        Returns:
            List[Field]: A list of Field instances representing the vector layer view's fields.
            
        Raises:
            ApiRequestError: If the API request fails.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> fields = view.get_fields()
        """
        return super().get_fields()
    

    def get_field(self, field_id: int) -> 'Field':
        """
        Get a specific field by its ID or name.
        
        Args:
            field_id (int): The ID of the field to retrieve.

        Returns:
            Field: The requested field instance.
            
        Raises:
            NotFoundError: If the field with the specified ID is not found.
            ApiRequestError: If the API request fails.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> field = view.get_field(id=1)
        """
        return super().get_field(field_id)


    def get_field_by_name(self, name: str) -> 'Field':
        """
        Get a specific field by its name.
        
        Args:
            name (str, optional): The name of the field to retrieve.

        Returns:
            Field: The requested field instance.
            
        Raises:
            NotFoundError: If the field with the specified name is not found. 

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> field = view.get_field_by_name(name='test')
        """      
        return super().get_field_by_name(name)
    

    def calculate_field(self, target_field: str, expression: str, q: str = None, bbox: List[float] = None, 
                        bbox_srid: int = None, feature_ids: List = None, run_async: bool = True, 
                        user_id: int = None) -> Union['Task', Dict]:
        """
        Calculate values for a field based on an expression.
        
        Args:
            target_field (str): The field to calculate values for.
            expression (str): The expression to use for calculation.
            q (Optional[str]): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            bbox (Optional[List[float]]): Bounding box to filter features.
            bbox_srid (Optional[int]): Spatial reference ID for the bounding box.
            feature_ids (Optional[List[int]]): List of specific feature IDs to include.
            run_async (Optional[bool]): Whether to run the calculation asynchronously.
            user_id (Optional[int]): ID of the user running the calculation.
            
        Returns:
            Task | Dict: The task instance of the calculation operation or the api response if the run_async=False.
            
        Raises:
            ValidationError: If the calculation parameters are invalid.
            ApiRequestError: If the API request fails.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = view.calculate_field(target_field="target_field", 
            ...                         expression="expression", 
            ...                         q="name like 'my_layer'", 
            ...                         bbox=[10, 20, 30, 40], 
            ...                         bbox_srid=3857, 
            ...                         feature_ids=[1, 2, 3], 
            ...                         run_async=True)
        """
        return super().calculate_field(target_field, expression, q, bbox, bbox_srid, feature_ids, run_async, user_id)
    

    def get_features(self, **kwargs) -> Union[List['Feature'], int]:
        """
        Get features from the layer with optional filtering and pagination.
        
        Keyword Args:
            quant_factor (int): Quantization factor. This parameter is only used by topojson encoder and is ignored for other formats. Higher quantizaion value means higher geometry precision. default is 1000000.
            skip (int): Number of features to skip. default is 0.
            limit (int): Maximum number of features to return. default is 100.
            user_id (int): Specific user. privileges required.
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            skip_geometry (bool): Whether to exclude geometry data. default is False.
            return_count (bool): Whether to return total count. default is False.
            feature_ids (List[int]): comma separated list of feature ids which should be filtered.
            select_fields (str): comma separated field names which should be included to the result. default is "[ALL]".
            skip_fields (str): comma separated field names which should be excluded from the result.
            out_srid (int): srid (epsg code) of result features. e.g. 4326. default is 3857
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, length D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            q (str): query filter based on OGC CQL standard. e.g. Name LIKE '%GIS%' AND INTERSECTS(geometry, 'SRID=3857;POLYGON((4901948 2885079, 7049893 2885079, 7049893 4833901, 4901948 4833901, 4901948 2885079))').
            bbox (str): Bounding box to filter features by. e.g. [50.275, 35.1195, 51.4459, 36.0416].
            bbox_srid (int): srid (epsg code) of bbox. e.g. 4326. default is 3857.
                
        Returns:
            List[Feature] | int: A list of Feature instances or the features count if return_count is True.
            

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> layer = VectorLayerView(api=client, name="my_layer", layer_type=LayerType.Point)
            >>> features = layer.get_features(quant_factor=1000000, 
            ...                                 skip=0, 
            ...                                 limit=100, 
            ...                                 skip_geometry=False, 
            ...                                 return_count=False, 
            ...                                 select_fields="fclass, osm_id", 
            ...                                 out_srid=4326, 
            ...                                 bbox_srid=4326)
        """
        return super().get_features(**kwargs)
    

    def get_feature(self, feature_id: int, out_srid: int = Feature.BASE_SRID) -> 'Feature':
        """
        Get a specific feature by its ID.
        
        Args:
            feature_id (int): The ID of the feature to retrieve.
            out_srid (int, optional): Output spatial reference ID

        Returns:
            Feature: The requested feature instance.
            
        Raises:
            NotFoundError: If the feature with the specified ID is not found.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> feature = view.get_feature(id=1)
        """
        return super().get_feature(feature_id, out_srid)
    

    def create_feature(self, geojson: Dict) -> 'Feature':
        """
        Create a new feature in the layer.
        
        Args:
            geojson (dict): The feature data including properties and geometry.
            
        Returns:
            Feature: The newly created feature instance.
            
        Raises:
            ValidationError: If the feature data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> feature = view.create_feature(geojson=geojson)
        """
        return super().create_feature(geojson)


    def delete_features(self, q: str = None, bbox: List[float] = None, bbox_srid: int = None, feature_ids: List[int] = None, 
                        run_async: bool = True, user_id: int = None) -> Union['Task', Dict]:
        """
        Delete features from the layer based on specified criteria.
        
        Args:
            q (Optional[str]): Query to filter features to delete.
            bbox (Optional[List[float]]): Bounding box to filter features.
            bbox_srid (Optional[int]): Spatial reference ID for the bounding box.
            feature_ids (Optional[List[int]]): List of specific feature IDs to delete.
            run_async (Optional[bool]): Whether to run the deletion asynchronously.
            user_id (Optional[int]): ID of the user performing the deletion.
            
        Returns:
            Task | Dict: The task instance of the deletion operation or the api response if run_async=False.
            
        Raises:
            ValidationError: If the deletion parameters are invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.delete_features(q="name like 'my_layer'", 
            ...                         bbox=[10, 20, 30, 40],
            ...                         bbox_srid=3857, 
            ...                         feature_ids=[1, 2, 3], 
            ...                         run_async=True)
        """
        return super().delete_features(q, bbox, bbox_srid, feature_ids, run_async, user_id)


    def import_features(self,
                            file: 'File', 
                            input_geom_type: 'InputGeomType' = None, 
                            input_layer_name: str = None, 
                            input_dataset: str = None, 
                            user_id: int = None, 
                            input_srid: int = None, 
                            file_encoding: str = "utf-8", 
                            replace_domain_codes_by_values: bool = False, 
                            report_errors: bool = True) -> 'Task':
        """
        Import features from a file into the layer.
        
        Args:
            file (File): file object to import.
            input_geom_type (InputGeomType, optional): Type of geometry in the input file.
            input_layer_name (str, optional): Name of the layer in the input file.
            input_dataset (str, optional): Name of the dataset in the input file.
            user_id (int, optional): Specific user.privileges requied.
            input_srid (int, optional): Spatial reference ID of the input data.
            file_encoding (str, optional): Character encoding of the input file.
            replace_domain_codes_by_values (bool, optional): Whether to replace domain codes with values.
            report_errors (bool, optional): Whether to report import errors.
            
        Returns:
            Task: The task instance of the import operation.
            
        Raises:
            ValidationError: If the import parameters are invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file = client.get_file(uuid="12345678-1234-5678-1234-567812345678")
            >>> task = view.import_features(file=file, 
            ...                         input_geom_type=InputGeomType.POINT,
            ...                         input_layer="my_layer",
            ...                         input_dataset="my_dataset",
            ...                         input_srid=3857,
            ...                         file_encoding="utf-8",
            ...                         replace_domain_codes_by_values=False, 
            ...                         report_errors=True)
        """
        return super().import_features(file, input_geom_type, input_layer_name, input_dataset, user_id, 
                                       input_srid, file_encoding, replace_domain_codes_by_values, report_errors)


    def export_features(self, out_filename: str, out_format: 'FileOutputFormat', replace_domain_codes_by_values: bool = False, 
                        run_async: bool = True, bbox: List[float] = None, out_srid: int = None, zipped: bool = True, 
                        feature_ids: List[int] = None, bbox_srid: int = None, q: str = None, fields: List[str] = None) -> Union['Task', Dict]:
        """
        Export features from the layer to a file.
        
        Args:
            out_filename (str): Name of the output file.
            out_format (FileOutputFormat): Format of the output file (e.g., 'Shapefile', 'GPKG', 'GeoJSON', 'CSV', 'KML', 'DXF').
            replace_domain_codes_by_values (bool, optional): Whether to replace domain codes with values.
            run_async (bool, optional): Whether to run the export asynchronously.
            bbox (List[float], optional): Bounding box to filter features.
            out_srid (int, optional): Spatial reference ID for the output.
            zipped (bool, optional): Whether to compress the output file.
            feature_ids (List[int], optional): List of specific feature IDs to export.
            bbox_srid (int, optional): Spatial reference ID for the bounding box.
            q (str, optional): Query to filter features.
            fields (List[str], optional): List of fields to include in the export.
            
        Returns:
            Task | Dict: The task instance of the export operation or the api response if run_async=False.
            
        Raises:
            ValidationError: If the export parameters are invalid.
            ApiRequestError: If the API request fails.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = view.export_features(out_filename="output.shp", 
            ...                         out_format="shp",
            ...                         replace_domain_codes_by_values=False, 
            ...                         run_async=True, 
            ...                         bbox=[10, 20, 30, 40], 
            ...                         out_srid=3857, 
            ...                         zipped=True, 
            ...                         feature_ids=[1, 2, 3])
        """
        return super().export_features(out_filename, out_format, replace_domain_codes_by_values, run_async, 
                                       bbox, out_srid, zipped, feature_ids, bbox_srid, q, fields)

    
    def get_tile_pbf_url(self, x: int = '{x}', y: int = '{y}', z: int = '{z}') -> str:
        """
        Get a vector tile pbf url for the view.
        
        Args:
            x (int): X coordinate of the tile.
            y (int): Y coordinate of the tile.
            z (int): Zoom level of the tile.

        Returns:
            str: the vector tile url.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> tile = view.get_tile(x=10, y=20, z=1)
        """
        return super().get_tile_pbf_url(x, y, z)
    

    def get_tile_json(self) -> Dict:
        """
        Get the vector tile JSON configuration for the layer.
        
        Returns:
            Dict: The vector tile JSON configuration.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> tile_json = view.get_tile_json()
        """
        return super().get_tile_json()


    @property
    def settings(self) -> Dict:
        """
        Get the layer's settings.
        
        Returns:
            Dict: The layer settings.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> setting = view.settings
        """
        return super().settings


    def update_settings(self, settings: Dict) -> Dict:
        """
        Update the settings

        settings (Dict): settings dictionary

        Returns:
            Dict: updated settings

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> view1 = client.get_view(uuid="12345678-1234-5678-1234-567812345678")
            >>> view2 = client.get_view(uuid="12345678-1234-5678-1234-567812345678")
            >>> view1.update_settings(view2.settings)
        """
        return super().update_settings(settings) 


    def set_settings(self, **kwargs) -> Dict:
        """
        Set the settings of the Vector Layer.
        
        Keyword Args:
            title_field (str): The field to use as the title.
            domain_display_type (str): The type of domain display.
            allow_export (bool): Whether to allow export.
            editable (bool): Whether to allow editing.
            edit_geometry (bool): Whether to allow editing the geometry.
            editable_attributes (str): The attributes to allow editing.
            allow_insert (bool): Whether to allow inserting.
            allow_delete (bool): Whether to allow deleting.
            min_zoom (int): The minimum zoom level.
            max_zoom (int): The maximum zoom level.
            max_features (int): The maximum number of features.
            filter_features (bool): Whether to filter features.
            fields (List[str]): The fields to include in the layer.
            use_cache (bool): Whether to use caching.
            cache_until_zoom (int): The zoom level to cache until.
            
        Returns:
            Dict: The updated settings.
            
        Raises:
            ValidationError: If the settings data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.set_setting(title_field="name",
            ...                      domain_display_type="Value",
            ...                      allow_export=True,
            ...                      editable=True,
            ...                      edit_geometry=True,
            ...                      editable_attributes="[ALL]",
            ...                      allow_insert=True,
            ...                      allow_delete=True,
            ...                      min_zoom=0,
            ...                      max_zoom=24,
            ...                      max_features=65536,
            ...                      filter_features=True,
            ...                      fields=["id"],
            ...                      use_cache=True,
            ...                      cache_until_zoom=17)
        """
        return super().set_settings(**kwargs)
    

    def seed_cache(self, from_zoom: int = None, to_zoom: int = None, ignore_cache: bool = False, workers: int = 1, user_id: int = None) -> List['Task']:
        """
        Seed the cache for the view.
        
        Args:
            from_zoom (int, optional): The zoom level to start caching from.
            to_zoom (int, optional): The zoom level to stop caching at.
            ignore_cache (bool, optional): Whether to ignore the cache. default is False.
            workers (int, optional): The number of workers to use. default is 1.
            user_id (int, optional): specified user. privileges required.

        Returns:
            List[Task]: The task instance of the cache seeding operation.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = view.cache_seed(from_zoom=0, to_zoom=10, ignore_cache=False, workers=1)
        """
        return super().seed_cache(from_zoom, to_zoom, ignore_cache, workers, user_id)


    def clear_cache(self) -> None:
        """
        Clear the view's cache.
        
        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.clear_cache()
        """
        return super().clear_cache()


    @property
    def cache_size(self) -> int:
        """
        Get the size of the view's cache.
        
        Returns:
            int: The size of the view's cache.
            
        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.cache_size
        """
        return super().cache_size


    def update_stats(self) -> None:
        """
        Update the view's statistics.
        
        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.update_stats()
        """
        return super().update_stats()


    def prune_edited_areas(self) -> None:
        """
        Prune edited areas. This method eliminates edited areas when there are too many of them. Cache builder uses this edited areas for partial cache generating.
        
        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.prune_edited_areas()
        """
        return super().prune_edited_areas()


    def get_attachments(self, **kwargs) -> List['Attachment']:
        """
        Get the resouces attachments

        Keyword Args:
            element_id (str): the id of the element with attachment.
            search (str): search term for keyword-based searching among all textual fields.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            return_count (bool): Whether to return total count. default is False.

        Returns:
            List[Attachment] | int: A list of attachments instances or the total number of attachments.

        Raises:
            TypeError: if the resource type is not supported

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> view.get_attachments()
        """
        from .attachment import Attachment

        return Attachment.get_attachments(self.api, resource=self, **kwargs)
    
    
    def create_attachment(self, 
                            name: str, 
                            loc_x: int,
                            loc_y: int,
                            file: 'File',
                            feature: 'Feature' = None,
                            display_name: str = None, 
                            description: str = None) -> 'Attachment':
        """
        Create a new Attachment.

        Args:
            name (str): The name of the scene.
            loc_x (int): x parameter of the attachment location.
            loc_y (int): y parameter of the attachment location.
            file (File): the file object.
            feature (Feature, optional): the feature object.
            display_name (str, optional): The display name of the scene.
            description (str, optional): The description of the scene.

        Returns:
            Attachment: The newly created Attachment instance.

        Raises:
            ValidationError: If the Attachment data is invalid.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> file = client.get_files()[0]
            >>> view.create_attachment(name='test', loc_x=10, loc_y=10, file=file)
        """
        from .attachment import Attachment

        return Attachment.create_attachment(self.api,
                                            name=name,
                                            loc_x=loc_x,
                                            loc_y=loc_y,
                                            resource=self,
                                            file=file,
                                            feature=feature,
                                            display_name=display_name,
                                            description=description)
    

    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncVectorLayerView':
        """
        Switch to async version of the view instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncVectorLayerView: the async instance of the view.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.view import VectorLayerView
            >>> client = GeoboxClient()
            >>> view = VectorLayerView.get_view(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_view = view.to_async(async_client)
        """
        from .aio.view import AsyncVectorLayerView

        return AsyncVectorLayerView(api=async_client, uuid=self.uuid, layer_type=self.layer_type, data=self.data)


import pytest
from urllib.parse import urljoin

from geobox.vectorlayer import VectorLayer, LayerType
from geobox.field import Field, FieldType


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_init(api, mock_field_data, mock_vector_data,  field_type):
    """Test Field initialization."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data)
    assert field.data_type == field_type
    assert field.datatype == field_type
    assert field.data == mock_field_data
    assert field.endpoint == urljoin(field.layer.endpoint, f"fields/{mock_field_data['id']}/")
    # error
    with pytest.raises(ValueError, match="data_type must be a FieldType instance"):
        field = Field(layer,
                data_type='invalid_data_type',
                data=mock_field_data)


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_init_without_id(api, mock_field_data, mock_vector_data,  field_type):
    """Test Field initialization without field_id."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    del mock_field_data['id']
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data)
    with pytest.raises(AttributeError):
        field.id
    assert field.endpoint is None
    assert field.data_type == field_type
    assert field.data == mock_field_data


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_repr(api, mock_field_data, mock_vector_data,  field_type):
    """Test the repr method of Field object for different data types"""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data)
    assert repr(field) == f"Field(id={field.id}, name={field.name}, data_type={field.data_type})"


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_get_domain(api, mock_field_data, mock_vector_data,  field_type):
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data)
    assert field.domain is None


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_set_domain(api, mock_field_data, mock_vector_data,  field_type):
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data)
    field.domain = {'min': 1, 'max': 10, 'items': {'1': 'value1', '2': 'value2'}}
    assert field.domain == {'min': 1, 'max': 10, 'items': {'1': 'value1', '2': 'value2'}}


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_create_field(api, mock_field_data, mock_vector_data,  field_type):
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    api.post.return_value = {**mock_field_data, 'datatype': field_type.value}
    field = Field.create_field(api, layer, 'test', field_type, mock_field_data)
    assert isinstance(field, Field)
    del mock_field_data['domain']
    del mock_field_data['width']
    del mock_field_data['display_name']
    del mock_field_data['description']
    api.post.assert_called_once_with(f"vectorLayers/{mock_vector_data['uuid']}/fields/", mock_field_data)


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_save_new_field(api, mock_field_data, mock_vector_data,  field_type):
    """Test saving a new field."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    id = mock_field_data['id']
    del mock_field_data['id']
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    with pytest.raises(AttributeError):
        field.id
    
    mock_field_data['id'] = id
    api.post.return_value = mock_field_data

    field.save()

    expected_data = {'name': 'test', 'datatype': field_type.value, 'hyperlink': False}

    api.post.assert_called_once_with(urljoin(field.layer.endpoint, 'fields/'), expected_data)
    assert field.id == mock_field_data["id"]
    assert field.data == mock_field_data
    assert field.endpoint == urljoin(field.layer.endpoint, f'fields/{mock_field_data["id"]}/')


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_save_existing_field(api, mock_field_data, mock_vector_data,  field_type):
    """Test saving an existing field."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    field.name = "new_osm_id"
    field.data["display_name"] = "New OSM ID"
    field.data["description"] = "New OSM ID field"
    field.save()

    expected_data = {
        "name": "new_osm_id",
        "datatype": field_type.value,
        "display_name": "New OSM ID",
        "description": "New OSM ID field",
        'hyperlink': False
    }

    api.put.assert_called_once_with(field.endpoint, expected_data)
    assert field.id == mock_field_data['id']
    assert field.name == "new_osm_id"
    assert field.data["display_name"] == "New OSM ID"
    assert field.data["description"] == "New OSM ID field"


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_delete(api, mock_field_data, mock_vector_data,  field_type):
    """Test deleting a field."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    endpoint = field.endpoint
    field.delete()
    api.delete.assert_called_once_with(endpoint)


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_update(api, mock_field_data, mock_vector_data,  field_type):
    """Test updating field properties."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    update_data = {
        "name": "updated_osm_id",
        "display_name": "Updated OSM ID",
        "description": "Updated OSM ID field"
    }
    
    api.put.return_value = {**field.data, **update_data}
    
    response = field.update(**update_data)
    
    expected_data = {'name': 'updated_osm_id', 'display_name': 'Updated OSM ID', 'description': 'Updated OSM ID field'}
    
    api.put.assert_called_once_with(field.endpoint, expected_data)
    assert response == {**field.data, **update_data}
    assert field.data == {**field.data, **update_data}


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_get_field_unique_values(api, mock_field_data, mock_vector_data,  field_type):
    """Test getting unique values of a field."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    expected_response = {"values": ["123456", "789012", "345678"]}
    api.get.return_value = expected_response
    
    response = field.get_field_unique_values()
    
    api.get.assert_called_once_with(urljoin(field.endpoint, 'distinct/'))
    assert response == expected_response


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_get_field_unique_values_numbers(api, mock_field_data, mock_vector_data,  field_type):
    """Test getting unique values count of a field."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    expected_response = {"count": 3}
    api.get.return_value = expected_response
    
    response = field.get_field_unique_values_numbers()
    
    api.get.assert_called_once_with(urljoin(field.endpoint, 'distinctCount/'))
    assert response == expected_response


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_get_field_statistic(api, mock_field_data, mock_vector_data,  field_type):
    """Test getting field statistics."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    expected_response = {"stat": 42}
    api.get.return_value = expected_response
    
    response = field.get_field_statistic("min")
    
    api.get.assert_called_once_with(urljoin(field.endpoint, 'stats/?func_type=min'))
    assert response == expected_response


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_update_domain(api, mock_field_data, mock_vector_data,  field_type):
    """Test updating the domain."""
    mock_field_data['datatype'] = field_type.value
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    range_d = {'min': 1, 'max': 10}
    list_d = {'1': 'value1', '2': 'value2'}
    mock_field_data['domain'] = {'min': 1, 'max': 10, 'items': {'1': 'value1', '2': 'value2'}}
    api.put.return_value = mock_field_data
    result = field.update_domain(range_domain = range_d, list_domain=list_d)
    assert result == {'min': 1, 'max': 10, 'items': {'1': 'value1', '2': 'value2'}}
    expected_response = {'name': 'test', 'datatype': field_type.value, 'domain': {'min': 1, 'max': 10, 'items': {'1': 'value1', '2': 'value2'}}, 'hyperlink': False}
    api.put.assert_called_once_with(field.endpoint, expected_response)


@pytest.mark.parametrize("field_type", [FieldType.Integer, FieldType.Float, FieldType.String])
def test_to_async(api, async_api, mock_vector_data, mock_field_data, field_type):
    layer = VectorLayer(api, uuid=mock_vector_data['uuid'], layer_type=LayerType.Polygon, data=mock_vector_data)
    field = Field(layer,
                    data_type=field_type,
                    data=mock_field_data.copy())
    async_instance = field.to_async(async_api)
    assert async_instance.api == async_api  


import pytest

from geobox.usage import Usage, UsageScale, UsageParam
from geobox.user import User
from geobox.apikey import ApiKey


def test_init(api, mock_user_data):
    user = User(api, mock_user_data['1']['id'], mock_user_data['1'])
    usage = Usage(api, user)
    assert usage.user == user
    assert usage.api == api


def test_repr(api, mock_user_data):
    user = User(api, mock_user_data['1']['id'], mock_user_data['1'])
    usage = Usage(api, user)
    assert repr(usage) == f"Usage(user={usage.user})"


def test_get_api_usage(api, mock_user_data, mock_apikey_data):
    data = [['2025-07-07T00:00:00', 21],
            ['2025-07-08T00:00:00', 267],
            ['2025-07-09T00:00:00', 1076],
            ['2025-07-11T00:00:00', 57],
            ['2025-07-12T00:00:00', 79]]
    api.get.return_value = data
    # user input
    user = User(api, mock_user_data['1']['id'], mock_user_data['1'])
    usage = Usage.get_api_usage(api, 
                               resource=user, 
                               scale=UsageScale.Day, 
                               param=UsageParam.Calls, 
                               days_before_now=5)
    assert usage == data
    api.get.assert_called_once_with('usage/api?eid=1&scale=day&param=calls&days_before_now=5')
    # apikey input
    api.reset_mock()
    apikey = ApiKey(api, mock_apikey_data['id'], mock_apikey_data)
    usage = Usage.get_api_usage(api, 
                               resource=apikey, 
                               scale=UsageScale.Day, 
                               param=UsageParam.Calls, 
                               days_before_now=5)
    assert usage == data
    api.get.assert_called_once_with('usage/api?eid=4t1dhyR1VHqzJDRY5ycO-daxVe-o1d8I-XzLriEHmiQ&scale=day&param=calls&days_before_now=5')
    
    # errors
    with pytest.raises(ValueError):
        Usage.get_api_usage(api, 
                            resource=apikey, 
                            scale=UsageScale.Day, 
                            param=UsageParam.Calls)

    with pytest.raises(ValueError):
        Usage.get_api_usage(api, 
                            resource='test', 
                            scale=UsageScale.Day, 
                            param=UsageParam.Calls,
                            days_before_now=5)


def test_get_process_usage(api):
    api.get.return_value = 23.458457
    process_usage = Usage.get_process_usage(api, days_before_now=5)
    assert process_usage == 23.458457

    # error
    with pytest.raises(ValueError):
        Usage.get_process_usage(api)


def test_get_usage_summary(api):
    data = {'available_storage': 396210733056,
                'used_storage': 1229062883,
                'max_daily_api_calls': 100000000,
                'daily_api_calls': 79,
                'max_monthly_api_calls': 3000000000,
                'monthly_api_calls': 2865,
                'max_daily_traffic': 10737418240,
                'daily_traffic': 251705,
                'max_monthly_traffic': 322122547200,
                'monthly_traffic': 50359396,
                'max_daily_process': 576,
                'daily_process': 0,
                'max_monthly_process': 17280,
                'monthly_process': 23,
                'remaining_days': None}
    api.get.return_value = data
    usage_summary = Usage.get_usage_summary(api)
    assert usage_summary == data


def test_update_usage(api):
    data = {'available_storage': 0,
            'used_storage': 1229062883,
            'max_daily_api_calls': 0,
            'daily_api_calls': 82,
            'max_monthly_api_calls': 0,
            'monthly_api_calls': 2868,
            'max_daily_traffic': 0,
            'daily_traffic': 252385,
            'max_monthly_traffic': 0,
            'monthly_traffic': 50360076,
            'max_daily_process': 0,
            'daily_process': 0,
            'max_monthly_process': 0,
            'monthly_process': 23,
            'remaining_days': None}
    api.post.return_value = data
    update = Usage.update_usage(api)
    assert update == data


def test_to_async(api, async_api, mock_user_data):
    user = User(api, mock_user_data['1']['id'], mock_user_data['1'])
    usage = Usage(api, user)
    async_instance = usage.to_async(async_api)
    assert async_instance.api == async_api


import os
import mimetypes
import requests
import sys
from urllib.parse import urljoin, urlencode
from typing import Optional, Dict, List, Optional, Union, TYPE_CHECKING

from .base import AsyncBase
from .task import AsyncTask
from ..utils import clean_data, join_url_params

if TYPE_CHECKING:
    from . import AsyncGeoboxClient
    from .user import AsyncUser
    from ..api import GeoboxClient
    from ..raster import Raster


class AsyncRaster(AsyncBase):

    BASE_ENDPOINT: str = 'rasters/'

    def __init__(self,
        api: 'AsyncGeoboxClient',
        uuid: str,
        data: Optional[Dict] = {}):
        """
        Constructs all the necessary attributes for the Raster object.

        Args:
            api (AsyncGeoboxClient): The API instance.
            uuid (str): The UUID of the raster.
            data (Dict, optional): The raster data.
        """ 
        super().__init__(api, uuid=uuid, data=data)


    @classmethod
    async def get_rasters(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncRaster'], int]:
        """
        [async] Get all rasters.

        Args:
            api (AsyncGeoboxClient): The API instance.

        Keyword Args:
            terrain (bool): whether to get terrain rasters.
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.        
            search_fields (str): comma separated list of fields for searching
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): whether to return the total count of rasters. default is False.
            skip (int): number of rasters to skip. minimum is 0.
            limit (int): number of rasters to return. minimum is 1.
            user_id (int): user id to show the rasters of the user. privileges required.
            shared (bool): whether to return shared rasters. default is False.

        Returns:
            List[AsyncRaster] | int: A list of Raster objects or the total count of rasters.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     rasters = await AsyncRaster.get_rasters(client, terrain=True, q="name LIKE '%GIS%'")
            or  
            >>>     rasters = await client.get_rasters(terrain=True, q="name LIKE '%GIS%'")
        """
        params = {
            'terrain': kwargs.get('terrain', None),
            'f': 'json',
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 100),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: AsyncRaster(api, item['uuid'], item))
    


    @classmethod
    async def get_rasters_by_ids(cls, api: 'AsyncGeoboxClient', ids: List[str], user_id: int = None) -> List['AsyncRaster']:
        """
        [async] Get rasters by their IDs.

        Args:
            api (AsyncGeoboxClient): The API instance.
            ids (List[str]): The IDs of the rasters.
            user_id (int, optional): specific user. privileges required.

        Returns:
            List['AsyncRaster']: A list of Raster objects.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     rasters = await AsyncRaster.get_rasters_by_ids(client, ids=['123', '456'])
            or  
            >>>     rasters = await client.get_rasters_by_ids(ids=['123', '456'])
        """ 
        params = {
            'ids': ids,
            'user_id': user_id,
        }
        endpoint = urljoin(cls.BASE_ENDPOINT, 'get-rasters/')

        return await super()._get_list_by_ids(api, endpoint, params, factory_func=lambda api, item: AsyncRaster(api, item['uuid'], item))


    @classmethod
    async def get_raster(cls, api: 'AsyncGeoboxClient', uuid: str, user_id: int = None) -> 'AsyncRaster':
        """
        [async] Get a raster by its UUID.

        Args:
            api (AsyncGeoboxClient): The API instance.
            uuid (str): The UUID of the raster.
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncRaster: A Raster object.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     raster = await client.get_raster(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return await super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: AsyncRaster(api, item['uuid'], item))


    @classmethod
    async def get_raster_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> Union['AsyncRaster', None]:
        """
        [async] Get a raster by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the raster to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncRaster | None: returns the raster if a raster matches the given name, else None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster_by_name(client, name='test')
            or  
            >>>     raster = await client.get_raster_by_name(name='test')
        """
        rasters = await cls.get_rasters(api, q=f"name = '{name}'", user_id=user_id)
        if rasters and rasters[0].name == name:
            return rasters[0]
        else:
            return None
    
    
    async def update(self, **kwargs) -> None:
        """
        [async] Update the raster.

        Keyword Args:
            name (str): The name of the raster.
            display_name (str): The display name of the raster.
            description (str): The description of the raster.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.update(name="new_name")
        """ 
        params = {
            'name': kwargs.get('name'),
            'display_name': kwargs.get('display_name'),
            'description': kwargs.get('description')
        }
        return await super()._update(self.endpoint, params)
    

    async def delete(self) -> None:
        """
        [async] Delete the raster.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.delete()
        """
        await super()._delete(self.endpoint)
    

    @property
    def thumbnail(self) -> str:
        """
        Get the thumbnail of the raster.

        Returns:
            str: The url of the thumbnail.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster.thumbnail
        """
        return super()._thumbnail(format='')
    
    
    @property
    async def info(self) -> Dict:
        """
        [async] Get the info of the raster.

        Returns:
            Dict: The info of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.info
        """
        endpoint = urljoin(self.endpoint, 'info/')
        return await self.api.get(endpoint)
    

    async def get_statistics(self, indexes: str = None) -> Dict:
        """
        [async] Get the statistics of the raster.

        Args:
            indexes (str): list of comma separated band indexes. e.g. 1, 2, 3

        Returns:
            Dict: The statistics of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.get_statistics(indexes='1, 2, 3')
        """
        params = clean_data({
            'indexes': indexes,
        })
        query_string = urlencode(params)
        endpoint = urljoin(self.endpoint, f'statistics/?{query_string}')
        return await self.api.get(endpoint)
    

    async def get_point(self, lat: float, lng: float) -> Dict:
        """
        [async] Get the point of the raster.

        Args:
            lat (float): The latitude of the point. minimum is -90, maximum is 90.
            lng (float): The longitude of the point. minimum is -180, maximum is 180.

        Returns:
            Dict: The point of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.get_point(lat=60, lng=50)
        """
        if lat < -90 or lat > 90:
            raise ValueError("lat must be between -90 and 90")
        if lng < -180 or lng > 180:
            raise ValueError("lng must be between -180 and 180")
        
        params = clean_data({
            'lat': lat,
            'lng': lng,
        })
        query_string = urlencode(params)
        endpoint = urljoin(self.endpoint, f'point?{query_string}')
        return await self.api.get(endpoint)
    

    def _get_save_path(self, save_path: str = None) -> str:
        """
        Get the path where the file should be saved.

        Args:
            save_path (str, optional): The path to save the file.

        Returns:
            str: The path where the file is saved.
        
        Raises:
            ValueError: If save_path does not end with a '/'.
        """
        # If save_path is provided, check if it ends with a '/'
        if save_path and save_path.endswith('/'):
            return f'{save_path}'
        
        if save_path and not save_path.endswith('/'):
            raise ValueError("save_path must end with a '/'")
        
        return os.getcwd()
    
    
    def _get_file_name(self, response: requests.Response) -> str:
        """
        Get the file name from the response.

        Args:
            response (requests.Response): The response of the request.

        Returns:
            str: The file name 
        """
        if 'Content-Disposition' in response.headers and 'filename=' in response.headers['Content-Disposition']:
            file_name = response.headers['Content-Disposition'].split('filename=')[-1].strip().strip('"')

        else:
            content_type = response.headers.get("Content-Type", "")
            file_name = f'{self.name}{mimetypes.guess_extension(content_type.split(";")[0])}'

        return file_name


    def _create_progress_bar(self) -> 'tqdm':
        """Creates a progress bar for the task."""
        try:
            from tqdm.auto import tqdm
        except ImportError:
            from .api import logger
            logger.warning("[tqdm] extra is required to show the progress bar. install with: pip insatll geobox[tqdm]")
            return None

        return tqdm(unit="B", 
                        total=int(self.size), 
                        file=sys.stdout,
                        dynamic_ncols=True,
                        desc="Downloading",
                        unit_scale=True,
                        unit_divisor=1024, 
                        ascii=True
                )


    async def download(self, save_path: str = None, progress_bar: bool = True) -> str:
        """
        [async] Download the raster.

        Args:
            save_path (str, optional): Path where the file should be saved. 
                                    If not provided, it saves to the current working directory
                                    using the original filename and appropriate extension.
            progress_bar (bool, optional): Whether to show a progress bar. default: True
        
        Returns:
            str: The path to save the raster.

        Raises:
            ValueError: If file_uuid is not set
            OSError: If there are issues with file operations        

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.download(save_path="path/to/save/")
        """
        if not self.uuid:
            raise ValueError("Raster UUID is required to download the raster file")
        
        save_path = self._get_save_path(save_path)
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        
        endpoint = urljoin(self.api.base_url, f"{self.endpoint}download/")

        async with self.api.session.session.get(endpoint) as response:
            file_name = self._get_file_name(response)
            full_path = f"{save_path}/{file_name}"
            with open(full_path, 'wb') as f:
                pbar = self._create_progress_bar() if progress_bar else None
                async for chunk in response.content.iter_chunked(8192):
                    f.write(chunk)
                    if pbar:
                        pbar.update(len(chunk))
                        pbar.refresh()
                if pbar:
                    pbar.close()

        return os.path.abspath(full_path)


    async def get_content_file(self, save_path: str = None, progress_bar: bool = True) -> str: 
        """
        [async] Get Raster Content URL

        Args:
            save_path (str, optional): Path where the file should be saved. 
                                    If not provided, it saves to the current working directory
                                    using the original filename and appropriate extension.
            progress_bar (bool, optional): Whether to show a progress bar. default: True
        
        Returns:
            str: The path to save the raster.

        Raises:
            ValueError: If uuid is not set
            OSError: If there are issues with file operations   

        Examples:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster_tiff = await raste.get_content_file()
        """
        if not self.uuid:
            raise ValueError("Raster UUID is required to download the raster content")
        
        save_path = self._get_save_path(save_path)
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        
        endpoint = urljoin(self.api.base_url, f"{self.endpoint}content/")
        
        async with self.api.session.session.get(endpoint) as response:
            file_name = self._get_file_name(response)
            full_path = f"{save_path}/{file_name}"
            with open(full_path, 'wb') as f:
                pbar = self._create_progress_bar() if progress_bar else None
                async for chunk in response.content.iter_chunked(8192):
                    f.write(chunk)
                    if pbar:
                        pbar.update(len(chunk))
                        pbar.refresh()
                if pbar:
                    pbar.close()

        return os.path.abspath(full_path)


    def get_render_png_url(self, x: int, y: int, z: int, **kwargs) -> str:
        """
        Get the PNG URL of the raster.

        Args:
            x (int): The x coordinate of the tile.
            y (int): The y coordinate of the tile.
            z (int): The zoom level of the tile.

        Keyword Args:
            indexes (str, optional): list of comma separated band indexes to be rendered. e.g. 1, 2, 3
            nodata (int, optional)
            expression (str, optional): band math expression. e.g. b1*b2+b3
            rescale (List, optional): comma (',') separated Min,Max range. Can set multiple time for multiple bands.
            color_formula (str, optional): Color formula. e.g. gamma R 0.5
            colormap_name (str, optional)
            colormap (str, optional): JSON encoded custom Colormap. e.g. {"0": "#ff0000", "1": "#00ff00"} or [[[0, 100], "#ff0000"], [[100, 200], "#00ff00"]]

        Returns:
            str: The PNG Render URL of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster.get_tile_render_url(x=10, y=20, z=1)
        """
        params = clean_data({
            'indexes': kwargs.get('indexes'),
            'nodata': kwargs.get('nodata'),
            'expression': kwargs.get('expression'),
            'rescale': kwargs.get('rescale'),
            'color_formula': kwargs.get('color_formula'),
            'colormap_name': kwargs.get('colormap_name'),
            'colormap': kwargs.get('colormap')
        })
        query_string = urlencode(params)
        endpoint = f'{self.api.base_url}{self.endpoint}render/{z}/{x}/{y}.png'
        if query_string:
            endpoint = f'{endpoint}?{query_string}'

        if not self.api.access_token and self.api.apikey:
            endpoint = join_url_params(endpoint, {"apikey": self.api.apikey})
            
        return endpoint
    

    def get_tile_pbf_url(self, x: int = '{x}', y: int = '{y}', z: int = '{z}', indexes: str = None) -> str:
        """
        Get the URL of the tile.

        Args:
            x (int, optional): The x coordinate of the tile.
            y (int, optional): The y coordinate of the tile.
            z (int, optional): The zoom level of the tile.
            indexes (str, optional): list of comma separated band indexes to be rendered. e.g. 1, 2, 3

        Returns:
            str: The URL of the tile.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster.get_tile_pbf_url(x=10, y=20, z=1)
        """
        params = clean_data({
            'indexes': indexes
        })
        query_string = urlencode(params)
        endpoint = urljoin(self.api.base_url, f'{self.endpoint}tiles/{z}/{x}/{y}.pbf')
        endpoint = urljoin(endpoint, f'?{query_string}')

        if not self.api.access_token and self.api.apikey:
            endpoint = join_url_params(endpoint, {"apikey": self.api.apikey})

        return endpoint
    

    def get_tile_png_url(self, x: int = 'x', y: int = 'y', z: int = 'z') -> str:
        """
        Get the URL of the tile.

        Args:
            x (int, optional): The x coordinate of the tile.
            y (int, optional): The y coordinate of the tile.
            z (int, optional): The zoom level of the tile.

        Returns:
            str: The URL of the tile.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster.get_tile_png_url(x=10, y=20, z=1)
        """
        endpoint = f'{self.api.base_url}{self.endpoint}tiles/{z}/{x}/{y}.png'
            
        if not self.api.access_token and self.api.apikey:
            endpoint = f'{endpoint}?apikey={self.api.apikey}'

        return endpoint


    async def get_tile_json(self) -> Dict:
        """
        [async] Get the tile JSON of the raster.

        Returns:
            Dict: The tile JSON of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.get_tile_json()
        """
        endpoint = urljoin(self.endpoint, 'tilejson.json')
        return await self.api.get(endpoint)


    def wmts(self, scale: int = None) -> str:
        """
        Get the WMTS URL

        Args:
            scale (int, optional): The scale of the raster. values are: 1, 2

        Returns:
            str: the raster WMTS URL

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster.wmts(scale=1)
        """ 
        endpoint = urljoin(self.api.base_url, f'{self.endpoint}wmts/')
        if scale:
            endpoint = f"{endpoint}?scale={scale}"

        if not self.api.access_token and self.api.apikey:
            endpoint = join_url_params(endpoint, {"apikey": self.api.apikey})

        return endpoint


    @property
    async def settings(self) -> Dict:
        """
        [async] Get the settings of the raster.

        Returns:
            Dict: The settings of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.settings
        """
        return await super()._get_settings(self.endpoint)
    

    async def update_settings(self, settings: Dict) -> Dict:
        """
        [async] Update the settings

        settings (Dict): settings dictionary

        Returns:
            Dict: updated settings

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster1 = client.get_raster(uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster2 = client.get_raster(uuid="12345678-1234-5678-1234-567812345678")
            >>>     raster1.update_settings(raster2.settings)
        """
        return await super()._set_settings(self.endpoint, settings)  


    async def set_settings(self, **kwargs) -> None:
        """
        [async] Set the settings of the raster.

        Keyword Args:
            nodata (int): The nodata value of the raster.
            indexes (list[int]): The indexes of the raster.
            rescale (list[int]): The rescale of the raster.
            colormap_name (str): The colormap name of the raster.
            color_formula (str): The color formula of the raster.
            expression (str): The expression of the raster.
            exaggeraion (int): The exaggeraion of the raster.
            min_zoom (int): The min zoom of the raster.
            max_zoom (int): The max zoom of the raster.
            use_cache (bool): Whether to use cache of the raster.
            cache_until_zoom (int): The cache until zoom of the raster.


        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.set_settings(nodata=0, 
            ...                             indexes=[1], 
            ...                             rescale=[[0, 10000]], 
            ...                             colormap_name='gist_rainbow', 
            ...                             color_formula='Gamma R 0.5', 
            ...                             expression='b1 * 2', 
            ...                             exaggeraion=10, 
            ...                             min_zoom=0, 
            ...                             max_zoom=22, 
            ...                             use_cache=True, 
            ...                             cache_until_zoom=17)
        """
        visual_settings = {
            'nodata', 'indexes', 'rescale', 'colormap_name', 
            'color_formula', 'expression', 'exaggeraion'
        }
        tile_settings = {
            'min_zoom', 'max_zoom', 'use_cache', 'cache_until_zoom'
        }

        settings = await self.settings

        for key, value in kwargs.items():
            if key in visual_settings:
                settings['visual_settings'][key] = value
            elif key in tile_settings:
                settings['tile_settings'][key] = value


        return await super()._set_settings(self.endpoint, settings)


    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the raster with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to share the raster with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await raster.share(users=users)
        """
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the raster with specified users.

        Args:
            users (List[AsyncUser]): The list of user objects to unshare the raster with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search="John")
            >>>     await raster.unshare(users=users)
        """
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the raster is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[AsyncUser]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)


    async def seed_cache(self, from_zoom: int = None, to_zoom: int = None, extent: List[int] = None, workers: int = 1) -> List['AsyncTask']:
        """
        [async] Seed the cache of the raster.

        Args:
            from_zoom (int, optional): The from zoom of the raster.
            to_zoom (int, optional): The to zoom of the raster.
            extent (List[int], optional): The extent of the raster.
            workers (int, optional): The number of workers to use. default is 1.

        Returns:
            AsyncTask: The task of the seed cache.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     task = await raster.seed_cache(from_zoom=0, to_zoom=22, extent=[0, 0, 100, 100], workers=1)
        """
        data = {
            'from_zoom': from_zoom,
            'to_zoom': to_zoom,
            'extent': extent,
            'workers': workers
        }
        return await super()._seed_cache(self.endpoint, data)


    async def clear_cache(self) -> None:
        """
        [async] Clear the cache of the raster.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.clear_cache()
        """
        await super()._clear_cache(self.endpoint)
        

    @property
    async def cache_size(self) -> int:
        """
        [async] Get the size of the cache of the raster.

        Returns:
            int: The size of the cache of the raster.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await AsyncRaster.get_raster(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await raster.cache_size
        """
        return await super()._cache_size(self.endpoint)


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Raster':
        """
        Switch to sync version of the raster instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Raster: the sync instance of the raster.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.raster import AsyncRaster
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     raster = await AsyncRaster.get_raster(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_raster = raster.to_sync(client)
        """
        from ..raster import Raster

        return Raster(api=sync_client, uuid=self.uuid, data=self.data)


    async def profile(self,
        polyline: List,
        number_of_samples: int = 100,
        output_epsg: Optional[int] = None,
        include_distance: bool = True,
        treat_nodata_as_null: bool = True) -> Dict:
        """
        [async] Create a profile form a raster along a path

        Args:
            polyline (List): Path coordinates as [x, y] pairs. Use raster CRS unless `output_epsg` is provided.
            number_of_samples (int, optional): Number of samples along the path. default: 100.
            output_epsg (int, optional): EPSG code for output coordinates. If None, use raster CRS.
            include_distance (bool, optional): Include cumulative distance for each sample. default: True.
            treat_nodata_as_null (bool, optional): Treat NoData pixels as null values. default: True.

        Returns:
            Dict: the profile result as geojson

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     raster = await client.get_raster(uuid="12345678-1234-5678-1234-567812345678")
            >>>     task = await raster.profile(polyline=[[0, 0], [10, 10]], number_of_samples=200)
        """
        endpoint = f"{self.endpoint}profile/"

        data = clean_data({
            'polyline': polyline,
            'number_of_samples': number_of_samples,
            'output_epsg': output_epsg,
            'include_distance': include_distance,
            'treat_nodata_as_null': treat_nodata_as_null
        })

        response = await self.api.post(endpoint=endpoint, payload=data)
        return response
